// Ben Newton 
// Computer Graphics (COMP 770)
// Programming Assignment - Ray Tracing - 1
// 
// Code based on a ray tracer I built during a ray tracing class taught by Pete Shirley at 
// the University of Utah using code similar to that in his book Realistic Ray Tracing.

#include "SceneLoader.h"
#include <iostream>
#include "Vector3.h"
#include "Ray.h"
#include "ImageBuffer.h"
#include "ONB.h"
#include "Vector2.h"
#include <math.h>
#include <stdlib.h>

using namespace std;

#define IMAGE_WIDTH 800
#define IMAGE_HEIGHT 600
#define BACKGROUND_R 0.1f
#define BACKGROUND_G 0.1f
#define BACKGROUND_B 0.1f
#define SQRT_OF_NUM_SAMPLES 4

double max(double a, double b);
void jitter(Vector2* samples, int num_samples); //for Jittered sampling

int main(int argc, char* argv[])
{
	scene::Color background(BACKGROUND_R, BACKGROUND_G, BACKGROUND_B, 1.0f); //make this changeable.
	//load the scene
	scene::Scene scn = scene::loadScene("spheres1.scene.xml");
	//scene::Scene scn = scene::loadScene("spheres2.scene.xml");
	//scene::Scene scn = scene::loadScene("myspheres.scene.xml");
	//scene::Scene scn = scene::loadScene("test2.scene.xml");

	ImageBuffer ib(IMAGE_WIDTH, IMAGE_HEIGHT);
	
	//initialize camera
	Vector3 lookAt(scn.camera.lookAtX, scn.camera.lookAtY, scn.camera.lookAtZ);
	Vector3 center(scn.camera.x, scn.camera.y, -scn.camera.z); //changing sign on z for camera to make image match
	Vector3 gaze = unitVector(lookAt - center);
	Vector3 vup(scn.camera.upX, scn.camera.upY, scn.camera.upZ);
	
	float val = tan(scn.camera.fov*3.14159265359/180.0/2.0);
	float aspectRatio = (float)IMAGE_WIDTH/(float)IMAGE_HEIGHT;
	float left = -val;
	float right = val; 
	float bottom = -val/aspectRatio; 
	float top = val/aspectRatio; 
	float distance = 1.0;  
	ONB uvw;
	uvw.initFromWV( -gaze, vup);
	Vector3 corner = center + left*uvw.u + bottom*uvw.v - distance*uvw.w;
	Vector3 across = (right-left)*uvw.u;
	Vector3 up = (top-bottom)*uvw.v;

	for(int i=0; i< IMAGE_HEIGHT; i++)
	{
		for(int j=0; j< IMAGE_WIDTH; j++)
		{
			scene::Color sum(0.0f,0.0f,0.0f,0.0f);
			int numSamples = SQRT_OF_NUM_SAMPLES*SQRT_OF_NUM_SAMPLES;
			Vector2* vec = new Vector2[numSamples];
			//Jittered Super sampling
			jitter(vec,numSamples);

			for(int l=0; l<numSamples; l++)
			{
					float a = ((float)j+vec[l].x())/((float)(IMAGE_WIDTH)); 
					float b = ((float)i+vec[l].y())/((float)(IMAGE_HEIGHT));

					Vector3 origin = center;
					Vector3 target = corner + a*across + b*up;

					Ray r(origin, unitVector(target-origin)); //not sure this needs to be unit

					float tmax = 100000.0f;
					bool hit = false;
					scene::HitRecord rec;
					// loop over spheres
					for ( size_t k = 0; k < scn.spheres.size(); k++ )
					{
						if(scn.spheres[k].hit(r,.00001f, tmax, rec))
						{
							tmax = rec.t;
							hit = true;				
						}
					}	
					if(hit)
					{
						//compute shadow ray to see if this hit point is in the shadows.
						Vector3 light(scn.lights[0].x, scn.lights[0].y, scn.lights[0].z);
						Vector3 hitPoint = (r.origin() + (rec.t*r.direction()));
						Ray shadowRay(hitPoint, unitVector(light - hitPoint));
						scene::HitRecord rec2;
						bool hit2 = false;
						for ( size_t k = 0; k < scn.spheres.size(); k++ )
						{
							if(scn.spheres[k].hit(shadowRay,.00001f, 100000.0f, rec2))//use hitShadow instead.
							{
								hit2 = true;	
							}
						}	
						scene::Color ambient = scene::Color(scn.lights[0].ambient, scn.lights[0].ambient, scn.lights[0].ambient, scn.lights[0].ambient);
						if(!hit2) //nothing blocking light
						{
							//determine v the direction from hit point to "eye"/camera
							Vector3 v = -r.direction();
							//determine unit vector pointing to light
							Vector3 l = unitVector(light - (r.origin() + (rec.t*r.direction())));
							double lambertian = dot(rec.normal, l);
							Vector3 vl = v+l;
							Vector3 h = vl/vl.length();
							float nh = dot(rec.normal, h);
							//Lambertian shading + ambient light + phong specular shading
							//need to fix to remove ambient as multipler for phong
							sum = sum + (rec.color*(ambient + scn.lights[0].color*max(0,lambertian)))+scn.lights[0].color*rec.material.reflectance*pow(max(0.0f,nh),rec.material.phongExponent); //add support for multiple lights
						}
						else
						{
							//ambient shadow 
							sum = sum + rec.color*ambient; //or black?
						}
					}
					else
					{
						sum = sum + background;
					}
				
			}//end sampling loop
			
			scene::Color avg(sum.r/numSamples, sum.g/numSamples, sum.b/numSamples, sum.a/numSamples);
			ib.set(j, i, avg);
		}
	}

	ib.writeBMP("test.bmp");

	return 0;
}

//where does this belong?
double max(double a, double b)
{
	if(a > b)
	{
		return a;
	}
	else
	{
		return b;
	}
}

void jitter(Vector2* samples, int num_samples)
{
	int sqrtSamples = (int)(sqrt((double)num_samples));
	for(int i=0; i < sqrtSamples; i++)
	{
		for(int j=0; j<sqrtSamples; j++)
		{
			float x = ((double)i + /*drand48()*/rand()/RAND_MAX) / (double)sqrtSamples;
			float y = ((double)j + /*drand48()*/rand()/RAND_MAX) / (double)sqrtSamples;
			(samples[i*sqrtSamples +j]).setX(x);
			(samples[i*sqrtSamples +j]).setY(y);
		}
	}
}

