#ifndef _VECTOR3_H_
#define _VECTOR3_H_

#include <math.h>

class Vector3
{
public:
	Vector3();
	Vector3(double x, double y, double z);
	Vector3(const Vector3& v);

	double x() const;
	double y() const;
	double z() const;

	Vector3 operator-() const;

	double length() const;

	void setX(double x);
	void setY(double y);
	void setZ(double z);

	Vector3 operator+(const Vector3& left)const;
	Vector3 operator-(const Vector3& left)const;
	friend Vector3 operator/(const Vector3& vec, double scalar);
	friend Vector3 operator*(double scalar, const Vector3& vec);

	//public for now speed(?), may reconsider later.
	double e[3];
};

//does this work okay?
double dot(const Vector3& v1, const Vector3& v2);
Vector3 cross(const Vector3& v1, const Vector3& v2);

Vector3 unitVector(const Vector3& v);

inline Vector3::Vector3() {}

inline Vector3::Vector3(double x, double y, double z)
{ 
	e[0] = x;
	e[1] = y;
	e[2] = z;
}

inline Vector3::Vector3(const Vector3& v)
{
	e[0] = v.e[0];
	e[1] = v.e[1];
	e[2] = v.e[2];
}

inline double Vector3::x() const 
{
	return e[0];
}

inline double Vector3::y() const 
{
	return e[1];
}

inline double Vector3::z() const 
{
	return e[2];
}

//negate
inline Vector3 Vector3::operator-() const
{
	return Vector3(-e[0], -e[1], -e[2]);
}

inline double Vector3::length() const
{
	return sqrt(e[0]*e[0] + e[1]*e[1] + e[2]*e[2]); 
}

inline void Vector3::setX(double x)
{
	e[0] = x;
}

inline void Vector3::setY(double y)
{
	e[1] = y;
}

inline void Vector3::setZ(double z)
{
	e[2] = z;
}

inline double dot(const Vector3& v1, const Vector3& v2)
{
	return (v1.x()*v2.x()) + (v1.y()*v2.y()) + (v1.z()*v2.z());
}

inline Vector3 cross(const Vector3& v1, const Vector3& v2)
{
	Vector3 temp;
	temp.e[0] = v1.y() * v2.z() - v1.z() * v2.y();
	temp.e[1] = v1.z() * v2.x() - v1.x() * v2.z();
	temp.e[2] = v1.x() * v2.y() - v1.y() * v2.x();
	return temp;
}

inline Vector3 Vector3::operator+(const Vector3& left) const
{
	Vector3 temp;
	temp.e[0] = e[0] + left.e[0];
	temp.e[1] = e[1] + left.e[1];
	temp.e[2] = e[2] + left.e[2];
	return temp;
}

inline Vector3 Vector3::operator-(const Vector3& left) const
{
	Vector3 temp;
	temp.e[0] = e[0] - left.e[0];
	temp.e[1] = e[1] - left.e[1];
	temp.e[2] = e[2] - left.e[2];
	return temp;
}

inline Vector3 operator/(const Vector3& vec, double scalar)
{
	Vector3 temp;
	temp.e[0] = vec.e[0] / scalar;
	temp.e[1] = vec.e[1] / scalar;
	temp.e[2] = vec.e[2] / scalar;
	return temp;
}

inline Vector3 operator*(double scalar, const Vector3& vec)
{
	Vector3 temp;
	temp.e[0] = vec.e[0] * scalar;
	temp.e[1] = vec.e[1] * scalar;
	temp.e[2] = vec.e[2] * scalar;
	return temp;
}

inline Vector3 unitVector(const Vector3& v)
{
	double length = v.length();
	Vector3 temp;
	temp.e[0] = v.e[0] / length;
	temp.e[1] = v.e[1] / length;
	temp.e[2] = v.e[2] / length;
	return temp;
}

#endif //_VECTOR3_H_
